use strict;
use warnings;

sub cleanup_bas() {
	my ($src_file, $subst_file, $dest_file) = @_;

	open(my $file, '<', $subst_file) or die "Cannot open substitution file";
	my @subst = <$file>;
	close $file;

	undef $/; 	# Read whole file in one "slurp"

	open($file, '<', $src_file) or die "Cannot open input file";
	my $bas = <$file>;
	close($file);

	my $no;
	my $l;
	my %newNames;
	my @remarks;
	my @delLines;

	foreach (@subst) {
		s/[\r\n]+$//;			# Remove linefeed

		if ($_ =~ /^-([0-9]+)$/) {
			# Remember lines to delete
			push(@delLines, 0 + $1);
			next;
		}
		
		my ($a, $b) = split(',', $_, 2);
		$a = quotemeta($a);
		
		if ($a =~ /^[0-9]+$/) {
			# Remember REMarks to put them in later
			push(@remarks, [$a, $b]);
			next;
		}
		
		# Replace token
		#
		# Sanity check variable type
		if (substr($a, 0, 3) eq "var") {
			my $sa = substr($a, -1); my $sb = substr($b, -1);
			$sa = "" if ($sa =~ /[\d\w]/);
			$sb = "" if ($sb =~ /[\d\w]/);
			die "Variable types for $a/$b do not match" if ($sa ne $sb);
		}
		
		# Check we don't have double names
		die "Variable $b already exists" if (exists($newNames{$b}));
		$newNames{$b} = 1;
		
		$a .= '([^\w])';			# Must be followed by a non-word charcter (to match "procFun161" but not "procFun16165")	
		$b = '"' . quotemeta($b) . '$1"';	# Enclose in quotes so $1 substitution works
		$bas =~ s/$a/$b/gmee;
	}

	# Sort lists
	@delLines = sort { $b <=> $a } @delLines;
	@remarks = sort { $b->[0] <=> $a->[0] } @remarks;

	# Create single-line IFs if possible
	# 20975   IF KEYROW (1) = var0C88% THEN
	# 20980     EXIT i : END IF
	$bas =~ s/\n(\d+\s+IF .* THEN)\n\d+\s+([^:\n]+) : END IF/\n$1 $2/gm;

	# Now process the file line by line to better handle the line numbers
	my $basNew = "";
	my $line = 0;
	my $nextDel = pop(@delLines);

	for (split /^/, $bas) {
		my $lastLine = $line;
		my @tokens = split(' ', $_);
		$line = 0 + $tokens[0];
		my $line2 = 2 + $line;
		my $line3 = 3 + $line;
		my $lastLine2 = 2 + $lastLine;

		# Delete line if requested
		while (defined $nextDel && $line > $nextDel) {
			$nextDel = pop(@delLines);
		}
		if (defined $nextDel && $line == $nextDel) {
			$nextDel = pop(@delLines);
			next;
		}
		
		# Skip REMark lines
		if (/\d+\s+REMark/) {
			$basNew .= $_;
			next;
		}
		
		# Insert ":" line before DEFines
		if (/(\d+ DEFine)/) {
			$basNew .= ($line-1) . " :\n";
		}
		
		# Move SELECT commands to own line
		# 20235       ON var1838% = 0,-1 : var1268% = 0
		if (/(\d+)(\s+)(ON [^:]*) : ([^\n]+)/) {
			$basNew .= "$line$2$3\n";
			$_ = "$line2$2 $4\n";
		}

		# Move two END IFs to their own lines
		# 7065    var0FF8 = 180 : END IF : END IF
		if (/(\d+)(\s+)  ([^\n]+) : END IF : END IF/) {
			$basNew .= "$line$2  $3\n$line2$2 END IF\n$line3$2END IF\n";
			next;		
		}

		# Move one END IF to its own line
		# 2130    var0C58% = 0 : END IF
		if (/(\d+)(\s+) ([^\n]+) : END IF/) {
			$basNew .= "$line$2 $3\n$line2$2END IF\n";
			next;		
		}
		
		# Remove surrounding IF parentheses
		# 3505   IF (var0B28%(var0818%) = var0C30) THEN
		s/(\d+\s+IF )\((.*)\)( THEN)/$1$2$3/gm;

		# Remove surrounding assignment parentheses
		# 1105 var0AE8 = (yres% - var0AE0%)
		s/(\d+\s+[A-Za-z0-9_%\$]+ = )\(([^:]*)\)/$1$2/gm;
		
		# Remove space before function calls
		s/(= [\w\d_]+) (\(((?:(?>[^()]+)|(?2))*)\))/$1$2/gm;
		
		# Remove some superfluous brackets
		s/,(\(((?:(?>[^()]+)|(?1))*)\)),/,$2,/gm;
		s/,(\(((?:(?>[^()]+)|(?1))*)\))$/,$2/gm;
		
		# Reformat DEFines
		s/(DEFine (PROCedure|FuNction) [\w]+[%\$]?) \(([\w]+[%\$]?)\)/$1\($3\)/gm;
		s/(DEFine (PROCedure|FuNction) [\w]+[%\$]?) \(([\w]+[%\$]?),([\w]+[%\$]?)\)/$1\($3, $4\)/gm;
		s/(DEFine (PROCedure|FuNction) [\w]+[%\$]?) \(([\w]+[%\$]?),([\w]+[%\$]?),([\w]+[%\$]?)\)/$1\($3, $4, $5\)/gm;
		s/(DEFine (PROCedure|FuNction) [\w]+[%\$]?) \(([\w]+[%\$]?),([\w]+[%\$]?),([\w]+[%\$]?),([\w]+[%\$]?)\)/$1\($3, $4, $5, $6\)/gm;
		s/(DEFine (PROCedure|FuNction) [\w]+[%\$]?) \(([\w]+[%\$]?),([\w]+[%\$]?),([\w]+[%\$]?),([\w]+[%\$]?),([\w]+[%\$]?)\)/$1\($3, $4, $5, $6, $7\)/gm;
		s/(DEFine (PROCedure|FuNction) [\w]+[%\$]?) \(([\w]+[%\$]?),([\w]+[%\$]?),([\w]+[%\$]?),([\w]+[%\$]?),([\w]+[%\$]?),([\w]+[%\$]?)\)/$1\($3, $4, $5, $6, $7, $8\)/gm;

		$basNew .= $_;	
	}

	# Add REMarks in last step as we want to work on the result of the previous step
	$bas = $basNew;
	$basNew = "";
	my $nextRem = pop(@remarks);

	for (split /^/, $bas) {
		my @tokens = split(' ', $_);
		$line = 0 + $tokens[0];

		while (defined $nextRem && $line > $nextRem->[0]) {
			# Insert REMark between lines
			$basNew .= "$nextRem->[0] REMark $nextRem->[1]\n";
			$nextRem = pop(@remarks);
		}
		
		if (defined $nextRem && $line == $nextRem->[0]) {
			# Insert REMark at end of line
			s/[\r\n]+$//;			
			chomp($_);
			$basNew .= "$_: REMark $nextRem->[1]\n";
			$nextRem = pop(@remarks);
			next;
		}
		
		$basNew .= $_;	
	}	

	# Double the indents
	$basNew =~ s/\n(\d+) ( +)([^ ])/\n$1 $2$2$3/gm;

	open(my $fh, '>', $dest_file);
	print $fh $basNew;
	close $fh;
	print "Done\n";
}

1;